import requests

from jet_bridge_base import settings
from jet_bridge_base.configuration import configuration
from jet_bridge_base.logger import logger


def api_method_url(method):
    return '{}/{}'.format(settings.API_BASE_URL, method)


def is_token_activated(project_token):
    if not project_token:
        return False

    url = api_method_url('project_tokens/{}/'.format(project_token))
    headers = {
        'User-Agent': '{} v{}'.format(configuration.get_type(), configuration.get_version())
    }

    r = requests.request('GET', url, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        return False

    result = r.json()

    return bool(result.get('activated'))


def project_auth(token, project_token, permission=None):
    if not project_token:
        return {
            'result': False
        }

    url = api_method_url('project_auth/')
    data = {
        'project_token': project_token,
        'token': token
    }
    headers = {
        'User-Agent': '{} v{}'.format(configuration.get_type(), configuration.get_version())
    }

    if permission:
        data.update(permission)

    r = requests.request('POST', url, data=data, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        logger.error('Project Auth request error: %d %s', r.status_code, r.reason)
        return {
            'result': False
        }

    result = r.json()

    if result.get('access_disabled'):
        return {
            'result': False,
            'warning': result.get('warning')
        }

    return {
        'result': True,
        'warning': result.get('warning')
    }



from sqlalchemy import inspect
from sqlalchemy.orm.base import ONETOMANY

from jet_bridge_base.db import get_mapped_base
from jet_bridge_base.models import data_types
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.serializers.model_description import ModelDescriptionSerializer
from jet_bridge_base.utils.common import merge
from jet_bridge_base.utils.db_types import map_data_type
from jet_bridge_base.views.base.api import APIView


class ModelDescriptionView(APIView):
    serializer_class = ModelDescriptionSerializer
    permission_classes = (HasProjectPermissions,)

    def get_queryset(self):
        non_editable = ['id']
        hidden = ['__jet__token']
        MappedBase = get_mapped_base(self.request)

        def map_column(column):
            params = {}

            try:
                data_type = map_data_type(column.type)
            except:
                data_type = 'NullType'

            if column.foreign_keys:
                foreign_key = next(iter(column.foreign_keys))
                data_type = data_types.FOREIGN_KEY
                params['related_model'] = {
                    'model': foreign_key.column.table.name
                }

            result = {
                'name': column.name,
                'db_column': column.name,
                'field': data_type,
                'filterable': True,
                'null': column.nullable,
                'editable': column.name not in non_editable,
                'params': params
            }

            if column.default is not None:
                result['default_type'] = 'value'
                result['default_value'] = column.default

            return result

        # def map_relation(relation):
        #     field = None
        #
        #     if relation.direction == ONETOMANY:
        #         field = 'ManyToOneRel'
        #
        #     return {
        #         'name': relation.key,
        #         'related_model': {
        #             'model': relation.table.name
        #         },
        #         'field': field,
        #         'related_model_field': relation.primaryjoin.right.name,
        #         'through': None
        #     }
        #
        # def table_relations(mapper):
        #     return list(map(map_relation, filter(lambda x: x.direction == ONETOMANY and hasattr(x, 'table'), mapper.relationships)))
        #
        # def table_m2m_relations(mapper):
        #     result = []
        #     name = mapper.selectable.name
        #
        #     for relation in mapper.relationships:
        #         if relation.direction != ONETOMANY or not hasattr(relation, 'table'):
        #             continue
        #
        #         m2m_relationships = relation.mapper.relationships.values()
        #
        #         if len(m2m_relationships) != 2:
        #             continue
        #
        #         if len(relation.table.columns) > 5:
        #             continue
        #
        #         self_relationship = m2m_relationships[1] if m2m_relationships[1].table.name == name else \
        #         m2m_relationships[0]
        #         other_relationship = m2m_relationships[0] if self_relationship == m2m_relationships[1] else \
        #         m2m_relationships[1]
        #
        #         result.append({
        #             'name': 'M2M {} {}'.format(self_relationship.table.name, other_relationship.table.name),
        #             'related_model': {
        #                 'model': other_relationship.table.name
        #             },
        #             'field': 'ManyToManyField',
        #             'related_model_field': self_relationship.table.name,
        #             'through': {'model': relation.table.name}
        #         })
        #
        #     return result

        def map_table(cls):
            mapper = inspect(cls)
            name = mapper.selectable.name

            from jet_bridge_base.configuration import configuration
            additional = configuration.get_model_description(name)

            result = {
                'model': name,
                'db_table': name,
                'fields': list(map(map_column, mapper.columns)),
                'hidden': name in hidden or name in configuration.get_hidden_model_description(),
                # 'relations': table_relations(mapper) + table_m2m_relations(mapper),
                'primary_key_field': mapper.primary_key[0].name
            }

            if additional:
                result = merge(result, additional)

            return result

        return list(map(map_table, MappedBase.classes))

    def get(self, *args, **kwargs):
        queryset = self.get_queryset()
        serializer = self.serializer_class(instance=queryset, many=True)
        return JSONResponse(serializer.representation_data)



from jet_bridge_base.settings import set_settings


class Configuration(object):

    def get_type(self):
        pass

    def get_version(self):
        pass

    def get_model_description(self, db_table):
        pass

    def get_hidden_model_description(self):
        return []

    def get_settings(self):
        pass

    def on_model_pre_create(self, model, pk):
        pass

    def on_model_post_create(self, model, instance):
        pass

    def on_model_pre_update(self, model, instance):
        pass

    def on_model_post_update(self, model, instance):
        pass

    def on_model_pre_delete(self, model, instance):
        pass

    def on_model_post_delete(self, model, instance):
        pass

    def media_get_available_name(self, path):
        pass

    def media_exists(self, path):
        pass

    def media_listdir(self, path):
        pass

    def media_get_modified_time(self, path):
        pass

    def media_open(self, path, mode='rb'):
        pass

    def media_save(self, path, content):
        pass

    def media_delete(self, path):
        pass

    def media_url(self, path, request):
        pass


configuration = Configuration()


def set_configuration(new_configuration):
    global configuration
    configuration = new_configuration
    set_settings(configuration.get_settings())



from jet_bridge_base.responses.json import JSONResponse


class ListAPIViewMixin(object):

    def list(self, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        paginate = not self.request.get_argument('_no_pagination', False)
        page = self.paginate_queryset(queryset) if paginate else None
        if page is not None:
            serializer = self.get_serializer(instance=page, many=True)
            return self.get_paginated_response(serializer.representation_data)

        if queryset._limit is None:
            queryset = queryset.limit(10000)

        serializer = self.get_serializer(instance=queryset, many=True)
        data = serializer.representation_data
        return JSONResponse(data)



from sqlalchemy import text
from sqlalchemy.exc import SQLAlchemyError

from jet_bridge_base import fields
from jet_bridge_base.db import create_session
from jet_bridge_base.exceptions.sql import SqlError
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.fields.sql_params import SqlParamsSerializers
from jet_bridge_base.serializers.serializer import Serializer


class SqlSerializer(Serializer):
    query = fields.CharField()
    timezone = fields.CharField(required=False)
    params = SqlParamsSerializers(required=False)

    def validate_query(self, value):
        forbidden = ['insert', 'update', 'delete', 'grant', 'show']
        for i in range(len(forbidden)):
            forbidden.append('({}'.format(forbidden[i]))
        if any(map(lambda x: ' {} '.format(value.lower()).find(' {} '.format(x)) != -1, forbidden)):
            raise ValidationError('forbidden query')

        i = 0
        while value.find('%s') != -1:
            value = value.replace('%s', ':param_{}'.format(i), 1)
            i += 1

        return value

    def execute(self, data):
        request = self.context.get('request')
        session = create_session(request)

        query = data['query']
        params = data.get('params', [])

        if 'timezone' in data:
            try:
                session.execute('SET TIME ZONE :tz', {'tz': data['timezone']})
            except SQLAlchemyError:
                pass

        try:
            result = session.execute(
                text(query),
                params
            )

            rows = list(map(lambda x: x.itervalues(), result))

            def map_column(x):
                if x == '?column?':
                    return
                return x

            return {'data': rows, 'columns': map(map_column, result.keys())}
        except (SQLAlchemyError, TypeError) as e:
            raise SqlError(e)
        finally:
            session.close()


class SqlsSerializer(Serializer):
    queries = SqlSerializer(many=True)

    def execute(self, data):
        serializer = SqlSerializer(context=self.context)

        def map_query(query):
            try:
                return serializer.execute(query)
            except SqlError as e:
                return {'error': str(e.detail)}

        return map(map_query, data['queries'])
